/**
 * Full Duplex Serial adapter API definition.
 */
#ifndef __BufferedSerial
#define __BufferedSerial

/**
 * Defines buffer length.
 */
#define SERIAL_BUFFER_SIZE 2048 /* works only with my modified BufferedSerial! */


/**
 * Defines mode bits
 *   mode bit 0 = invert rx
 *   mode bit 1 = invert tx
 *   mode bit 2 = open-drain/source tx
 *   mode bit 3 = ignore tx echo on rx
 */
#define SERIAL_MODE_INVERT_RX 1
#define SERIAL_MODE_INVERT_TX 2
#define SERIAL_MODE_OPENDRAIN_TX 4
#define SERIAL_MODE_IGNORE_TX_ECHO 8

typedef struct {
    int rx_head;   // receive queue head
    int rx_tail;   // receive queue tail
    int tx_head;   // transmit queue head
    int tx_tail;   // transmit queue tail
    int rx_pin;    // recieve pin
    int tx_pin;    // transmit pin
    int mode;      // interface mode
    int ticks;     // clkfreq / baud
    void* rxBufferPtr;
    void* txBufferPtr;
    int buffer_size;
} tSerialStruct;

/**
 * start initializes and starts native assembly driver in a cog. 
 * @param rxpin is pin number for receive input
 * @param txpin is pin number for transmit output
 * @param mode is interface mode
 * @param baudrate is frequency of bits ... 115200, 57600, etc...
 * @returns non-zero on success
 */
int SerialStart(int rxpin, int txpin, int mode, int baudrate);
void SerialStop(void);
void SerialRxFlush(void);
int SerialRxCheck(void);
int SerialRxTime(int ms);
int SerialRx(void);
int SerialTx(int txChar);

#endif 
